//---------------------------------------------------------------------------------------------------------------------
//
// \copyright NATIVE INSTRUMENTS GmbH, 2016. All rights reserved.
//
//---------------------------------------------------------------------------------------------------------------------

#pragma once

#include <memory>

namespace ni { namespace dsp {

//! An array containing two constant float buffers
using stereo_in   = const float* const [2];
  
//! An array containing two float buffers
using stereo_out  = float* const [2];

  
//! \brief The Stem Mastering Chain is composed of a Compressor followed by a Limiter to create the finalized sound 
//! required by the Stem File format. Both Compressor and Limiter can be turned on or off individually 
//! (and are bypassed in the off state).
//! The Compressor's key input signal can optionally be highpass filtered. The highpass filter can be
//! bypassed by setting its cutoff frequency lower than 20Hz. 
//! The Stem Mastering Chain has a 32 sample lookahead @ 44.1kHz sample rate, and the number of lookahead samples is 
//! scaled by sample rate.
//! All parameters are set in natural values (dB, sedonds, etc.). 
//! Level and Gain Reduction meters are optional and can be activated or deactivated. They return display values in dB.
class stem_chain final
{
public:
  
  //! \brief Creates a Stem Mastering Chain with the specified sample rate.
  stem_chain( float sample_rate = 44100.f );
  
  
  //! \brief Process a two channel input signal for the specified amount of frames.
  //! The resulting audio samples are written into stereo_out.
  void process( stereo_in&, size_t frames, stereo_out& );
  
  //! \brief Computes the gain reduction curve delayed by the Stem Mastering Chain's lookahead size for 
  //! the specified amount of frames. 
  //! The out_gain_reduction parameter contains the computed values.
  //! Note that if this gain reduction curve is to be applied to the input signal, the input signal needs 
  //! to be delayed by the lookahead size to align it with the curve.
  //! The number of lookahead samples can be retrieved from lookahead_size().
  void gain_reduction( stereo_in&, size_t frames, float* out_gain_reduction );
  
  //! \brief Returns the limiter's lookahead size in samples.
  //! Lookahead size is 32 samples @ 44.1kHz and gets scaled by sample rate.
  int lookahead_size() const;

  
  //! \brief Sets the desired sample rate.
  void sample_rate( float );
  
  //! \brief Returns the current sample rate.
  float sample_rate() const;

  
  // Compressor Interface
  
  //! \brief Activates or deactivates/bypasses the Compressor.
  void compressor_active( bool );
  
  //! \brief Returns true if the Compressor is activated.
  bool compressor_active() const;
  
  
  //! \brief Set the Compressor's input gain in dB clamped to [-12.0f, 12.0f].
  void compressor_input_gain( float );
  
  //! \brief Returns the Compressor's input gain in dB.
  float compressor_input_gain() const;
  
  
  //! \brief Set the Compressor's output gain in dB clamped to [-12.0f, 12.0f].
  void compressor_output_gain( float );
  
  //! \brief Returns the Compressor's output gain in dB.
  float compressor_output_gain() const;

  
  //! \brief Set the Compressor's threshold in dB clamped to [-40.0f, 0.0f].
  void compressor_threshold( float );
  
  //! \brief Returns the Compressor's threshold in dB.
  float compressor_threshold() const;

  
  //! \brief Set the Compressor's dry/wet ratio in percent.
  //! The parameter specifies the wetness in percent and is clamped to [0.0f, 100.0f] percent.
  void compressor_wet_dry_gain( float );
  
  //! \brief Returns the Compressor's wetness in percent.
  float compressor_wet_dry_gain() const;

  
  //! \brief Set the Compressor's ratio.
  //! Valid ratio values are [1.5f (1:1.5), 2.0f (1:2), 3.0f (1:3), 4.0f (1:4), 5.0f (1:5), 10.0f (1:10)].
  //! If a non-valid ratio value is entered, the ratio will be set to the next value above with a max value of 10.f.
  void compressor_ratio( float );
  
  //! \brief Returns the Compressor's ratio.
  float compressor_ratio() const;
  
  
  //! \brief Set the Compressor's attack time in seconds clamped to [0.0001f, 0.03f].
  void compressor_attack( float );
  
  //! \brief Returns the Compressor's attack time in seconds.
  float compressor_attack() const;
  
  
  //! \brief Sets the Compressor's release time in seconds clamped to [0.1f, 1.6f].
  void compressor_release( float );
  
  //! \brief Returns the Compressor's release time in seconds.
  float compressor_release() const;
  
  
  //! \brief Sets the Compressor's key input highpass filter cutoff frequency in Hz.
  //! Valid values are in the range [20.0f, 500.0f].
  //! A value below 20.0f will effectively disable the highpass filter.
  //! Highpass filter is applied to the sidechain signal fed into Compressor's gain computer. That is, 
  //! it affects the gain reduction curve, but is not applied directly to the Compressor's audio output.
  void compressor_hp_cutoff( float );
  
  //! \brief Returns the Compressor's highpass filter cutoff frequency in Hz.
  //! A value of 0.f indicates that the highpass filter is disabled.
  float compressor_hp_cutoff() const;
  
  
  // Limiter Interface
  
  //! \brief Activates or deactivates/bypasses the Limiter.
  void limiter_active( bool );
  
  //! \brief Returns true if the Limiter is activated.
  bool limiter_active() const;
  
  
  //! \brief Sets the Limiter's threshold in dB clamped to [-40.0f, 0.0f].
  void limiter_threshold( float );
  
  //! \brief Returns the Limiter's threshold in dB.
  float limiter_threshold() const;
  
  
  //! \brief Sets the Limiter's release time in seconds clamped to [0.001f, 1.f].
  void limiter_release( float );
  
  //! \brief Returns the Limiter's release time in seconds.
  float limiter_release() const;
  
  
  //! \brief Sets the Limiter's ceiling in dB clamped to [-40.f, -0.001f].
  void limiter_ceiling( float );
  
  //! \brief Returns the Limiter's ceiling in dB.
  float limiter_ceiling() const;
  
  
  // Meters
  
  //! \brief Activate or deactivate the computation of the Level Meters.
  void meters_active( bool );

  //! \brief Returns true if the Level Meters are activated.
  bool meters_active() const;
  
  
  // Compressor Meters
  
  //! \brief Returns the Compressor's input RMS level in dB.
  //! ~3dB offset is added such that meter level is at 0dB for a 1kHz input sine wave. 
  //! Min supported meter range is -100dB.
  float compressor_input_level_rms() const;
  
  //! \brief Returns the Compressor's output RMS level in dB.
  //! ~3dB offset is added such that meter level is at 0dB for a 1kHz input sine wave. 
  //! Min supported meter range is -100dB.
  float compressor_output_level_rms() const;
  
  //! \brief Returns the Compressor's input peak level in dB. 
  //! Min supported meter range is -100dB.
  float compressor_input_level_peak() const;
  
  //! \brief Returns the Compressor's output peak level in dB.
  //! Min supported meter range is -100dB.
  float compressor_output_level_peak() const;
  
  //! \brief Returns the Compressor's gain reduction in positive dB values.
  float compressor_gain_reduction() const;

  
  // Limiter Meters
  
  //! \brief Returns the Limiter's input RMS level in dB.
  //! ~3dB offset is added such that meter level is at 0dB for a 1kHz input sine wave. 
  //! Min supported meter range is -100dB.
  float limiter_input_level_rms() const;
  
  //! \brief Returns the Limiter's output RMS level in dB. 
  //! ~3dB offset is added such that meter level is at 0dB for a 1kHz input sine wave. 
  //! Min supported meter range is -100dB.
  float limiter_output_level_rms() const;
  
  //! \brief Returns the Limiter's input peak level in dB.
  //! Min supported meter range is -100dB.
  float limiter_input_level_peak() const;
  
  //! \brief Returns the Limiter's output peak level in dB. 
  //! Min supported meter range is -100dB.
  float limiter_output_level_peak() const;
  
  //! \brief Returns the Limiter's gain reduction in positive dB values.
  float limiter_gain_reduction() const;
  
  
  ~stem_chain();
  
private:
  
  std::unique_ptr< class stem_chain_impl > m_impl;
};
  
} }
